from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys

from lcli.input.exctptions import InterruptedInputException
from lcli.input.prompt_toolkit import input_dialog, radiolist_dialog


class AskQuestions(object):
    _questions: dict

    def __init__(self, questions: dict):
        self._questions = questions

    def ask(self):
        result = {}
        errors = []
        for key in self._questions:
            answer = None
            try:
                question = self._questions[key]
                if question["type"] == "list":
                    answer = self.ask_radio(question)

                if question["type"] == "autocomplete":
                    answer = self.ask_autocompleter(question)

                if question["type"] == "input":
                    answer = self.ask_text(question)
            except Exception as e:
                errors.append(e)

            if answer is None:
                raise InterruptedInputException(
                    "Input process was stopped", result, errors
                )

            result[key] = answer
        return result

    @classmethod
    def ask_radio(cls, question):
        choices = question["choices"]
        asked_choices = []
        if type(choices) == list:
            for choice in choices:
                asked_choices.append((choice, choice))
        if type(choices) == dict:
            for choice in choices:
                asked_choices.append((choice, choices[choice]))

        if len(asked_choices) == 0:
            asked_choices = [("N/A", "N/A")]
        buttons = {}
        if "buttons" in question:
            buttons = question["buttons"]

        result = radiolist_dialog(
            values=asked_choices,
            title=question["message"],
            text=question["message"],
            buttons=buttons,
        ).run()
        return result

    @classmethod
    def ask_autocompleter(cls, question):
        choices_completer = WordCompleter(
            question["choices"], ignore_case=True, sentence=True
        )
        # Key bindings.
        bindings = KeyBindings()

        @bindings.add(Keys.Escape)
        @bindings.add("c-c")
        @bindings.add("c-q")
        def _(event):
            """Pressing Ctrl-Q or Ctrl-C will exit the user interface."""
            event.app.exit()

        return prompt(
            question["message"],
            completer=choices_completer,
            complete_while_typing=False,
            key_bindings=bindings,
        )

    def ask_text(self, question):
        app = input_dialog(title=question["message"], text=question["message"])
        result = app.run()

        return result



from lcli.app_mode.exceptions import AppModeException


class AppModeBase:
    from lcli.app import App

    _app: App

    def __init__(self, app: App) -> None:
        self._app = app

    def run(self) -> None:
        """Run application in current mode"""
        raise AppModeException(
            "Application mode not implemented or is wrong configured."
        )



import glob
import os

from lcli import __version__
from lcli.app import App
from lcli.tools.base import BaseTool


class Helper(BaseTool):
    """Application helper"""

    def __init__(self, app: App) -> None:
        super().__init__(app)

    def groups(self) -> list:
        """List all command groups"""
        available_groups = self._app.get_config_object().get_available_groups()
        result = []
        if len(available_groups) == 0:
            result.append("All")
        else:
            for group in available_groups:
                if group == "-g":
                    continue
                result.append(group)
        return result

    def commands(self):
        """List all command groups"""
        result = []
        available_commands = self._app.get_config_object().get_available_commands()
        if len(available_commands) == 0:
            result.append("All")
        for command in available_commands:
            result.append(command)
        return result

    def cache_clear(self):
        files = glob.glob(
            os.path.join(self._app.cache_directory_path, "*"), recursive=True
        )
        for f in files:
            try:
                os.remove(f)
            except OSError as e:
                print("Error: %s : %s" % (f, e.strerror))


def version():
    """Display application version"""
    print("Version: " + __version__)


def verbose():
    """Enable logging"""
    raise Exception("Verbose mode is not available for the moment")



import os
from cmd import Cmd

from pyfiglet import Figlet

from lcli.api.conroller import ControllerInterface
from lcli.app_mode.base import AppModeBase, AppModeException
from lcli.command.input import ParametersReader
from lcli.command.subprocess import CommandRunner
from lcli.config import Command, CommandArgs
from lcli.input.exctptions import InterruptedInputException
from lcli.input.prompt import AskQuestions
from lcli.tools.base import ToolsException


def run():
    """Run application in interactive mode for best productivity"""
    raise AppModeException("App mode unavailable for the moment")


class LcliPrompt(Cmd, AppModeBase, ControllerInterface):
    from lcli.app import App

    current_command = ""

    prompt_prefix = ""
    prompt = prompt_prefix + "> "
    intro = ""
    space_separator: str
    app_help_message: str

    doc_header = "Interactive cli commands (type help <topic>):"

    _RESET_WORD = "RESET"
    _CANCEL_WORD = "CANCEL"

    _COMMAND_MAPPING = {
        "/": "change_command",
        "//": "reset_command",
        "<": "remove_one_level_command",
        "///": "force_change_command",
        "z": "clear_screen",
    }

    def __init__(
        self, app: App, command_prefix="", completekey="tab", stdin=None, stdout=None
    ):
        super().__init__(completekey, stdin, stdout)
        AppModeBase.__init__(self, app)
        self.do_clear_screen()
        figlet = Figlet(
            width=1000, font=app.get_config_object().get("figlet_font", "digital")
        )
        command_parts = command_prefix.split(" ")
        if command_parts[0] == self._app.get_executable_name():
            command_parts = command_parts[1:]
        self.current_command = " ".join(command_parts)

        cli_code = app.get_app_code()
        self.prompt_prefix = self._app.__.red(cli_code)
        self.space_separator = self._app.get_config_object().get(
            "app_space_separator", "\n\n\n\n"
        )
        self.app_help_message = self._app.get_config_object().get(
            "app_help_message", ""
        )
        self.intro = (
            self.space_separator
            + figlet.renderText(str(cli_code).upper())
            + self.space_separator
            + self._app.__.red(
                self._app.get_config_object().get(
                    "app_welcome_message", "Welcome! Type ? to list commands"
                )
            )
        )
        self.doc_header = self._app.get_config_object().get(
            "app_doc_header", self.doc_header
        )
        self.refresh_prompt()

    def run(self) -> None:
        """Execute interactive shell for managing scripts"""
        self.cmdloop()

    def do_exit(self, inp):
        print(self._app.__.red("Bye"))
        return True

    def help_exit(self):
        print(self._app.__.red("\nExit the application. Shorthand: x, q or Ctrl-D.\n"))

    def refresh_prompt(self):
        self.prompt = self.prompt_prefix
        if self.current_command != "":
            self.prompt += " " + self._app.__.green(self.current_command)

        ignored_cli_info = self._app.get_config_object().get("cli_info_ignore", [])
        info_parts = []
        if self.current_command not in ignored_cli_info:
            info_parts = self._invoke_command(
                self.current_command, "cli_info", "list", default=[]
            )

        command_state = ""
        for part in info_parts:
            command_state += " (" + self._app.__.yellow(part) + ")"
        self.prompt += command_state
        self.prompt += "\n> "

    def do_clear_screen(self, command=""):
        """Clear the screen and move the cursor to the top"""
        os.system("cls" if os.name == "nt" else "clear")
        self.stdout.write(str(self.intro) + "\n")

    def do_reset_command(self, command=""):
        """Reset current command prefix. This will allow to run any of cli commands."""
        self.current_command = ""
        self.refresh_prompt()

    def do_remove_one_level_command(self, command=""):
        """Remove one level of the current command"""
        self.current_command = " ".join(self.current_command.split(" ")[0:-1])
        self.refresh_prompt()

    def do_force_change_command(self, command=""):
        """Force change interactive command prefix. This will allow running sub-commands from chosen group."""
        self.current_command = ""
        self.do_change_command("")

    def do_change_command(self, command=""):
        """Change interactive command prefix. This will allow running sub-commands from chosen group."""
        if command == "":
            path = []
            if self.current_command != "":
                path = self.current_command.split()

            available_buttons = {self._CANCEL_WORD: {"text": self._CANCEL_WORD}}
            if self.current_command != "":
                available_buttons[self._RESET_WORD] = {"text": self._RESET_WORD}

            names_in_path = self._app.get_commands().get_commands_names_in_path(
                path,
                available_groups=self._app.get_config_object().get_available_groups(),
            )
            names_in_path_list = [k for k in names_in_path]
            names_in_path_list.sort()
            if len(names_in_path_list) == 0:
                names_in_path_list = {self._CANCEL_WORD: "N/A"}

            question = {
                "type": "list",
                "name": "command",
                "message": "Choose a command: ",
                "buttons": available_buttons,
                "choices": names_in_path_list,
            }
            try:
                questions = AskQuestions({"command": question})
                answers = questions.ask()
            except InterruptedInputException:
                return

            if "command" in answers:
                if answers["command"] is None:
                    return

                if answers["command"] == self._RESET_WORD:
                    self.current_command = ""
                elif answers["command"] == self._CANCEL_WORD:
                    pass
                else:
                    separator = ""
                    if self.current_command != "":
                        separator = " "
                    self.current_command += separator + answers["command"]
            else:
                print("Empty selected command. Please try again ...")
                self.do_change_command(command)
        else:
            self.current_command = command

        self.refresh_prompt()

    def complete_change_command(self, text, line, begidx, endidx):
        path = []
        if line != "":
            path = line.split()
        available_commands = []
        for command in self._app.get_commands().get_commands_names_in_path(
            path[1:],
            available_groups=self._app.get_config_object().get_available_groups(),
        ):
            available_commands.append(command)
        print(available_commands)
        return [i for i in available_commands if i.startswith(text)]

    @classmethod
    def help_change_command(cls):
        """Change interactive command prefix. This will allow running sub-commands from chosen group."""
        print("Change interactive command prefix.")
        print("This will allow running sub-commands from chosen group.")

    def emptyline(self):
        if self.current_command != "help":
            command_parts = self.current_command.split(" ")
            if command_parts[0] == self._app.get_executable_name():
                command_parts = command_parts[1:]
            if len(command_parts) > 1:
                command_path = " ".join(command_parts[:-1])
                command_key = command_parts[-1]
            else:
                command_path = self.current_command
                command_key = ""

            returned_value = self._invoke_command(command_path, command_key)
            self._print_command_result(returned_value)
        else:
            self.do_help({})

    def do_help(self, arg):
        """Show help message for interactive mode."""
        super().do_help(arg)
        self.stdout.write("%s\n" % self.app_help_message)

    def default(self, inp):
        if inp == "x" or inp == "q":
            return self.do_exit(inp)
        command = inp
        returned_value = self._invoke_command(self.current_command, command)
        self._print_command_result(returned_value)
        self.refresh_prompt()

    def complete(self, text, state):
        """Return the next possible completion for 'text'.

        If a command has not been entered, then complete against command list.
        Otherwise try to call complete_<command> to get list of completions.
        """
        if state == 0:
            import readline

            origline = readline.get_line_buffer()
            line = origline.lstrip()
            stripped = len(origline) - len(line)
            begidx = readline.get_begidx() - stripped
            endidx = readline.get_endidx() - stripped

            cmd = ""
            if begidx > 0:
                cmd, args, foo = self.parseline(line)
                if cmd == "" or cmd is None:
                    compfunc = self.complete_command
                else:
                    try:
                        compfunc = getattr(self, "complete_" + cmd)
                    except AttributeError:
                        compfunc = self.complete_command
            else:
                compfunc = self.completenames

            if not hasattr(self, "complete_" + cmd) and self.current_command != "":
                compfunc = self.complete_command

            self.completion_matches = compfunc(text, line, begidx, endidx)
        try:
            return self.completion_matches[state]
        except IndexError:
            return None

    def completenames(self, text, *ignored):
        dotext = "do_" + text
        available_commands = [a[3:] for a in self.get_names() if a.startswith(dotext)]
        for cc in self._app.get_commands().get_commands_names_in_path(
            [], available_groups=self._app.get_config_object().get_available_groups()
        ):
            if cc.startswith(text):
                available_commands.append(cc)
        return available_commands

    def complete_command(self, text, line: str, begidx, endidx):
        parts = line.rstrip().split(" ")
        sub_command = " ".join(parts[1:-1])
        if self.current_command != "":
            current_command = self.current_command
        else:
            current_command = parts[0]
        cmd = current_command + " " + sub_command

        available_commands = self._invoke_command(cmd, "h", "list", default=[])
        return [i for i in available_commands if i.startswith(text)]

    def parseline(self, line):
        line = line.strip()
        parts = line.split()
        if not line:
            return None, None, line
        else:
            if parts[0] in self._COMMAND_MAPPING:
                line = self._COMMAND_MAPPING[parts[0]] + " " + " ".join(parts[1:])
        return super().parseline(line)

    do_EOF = do_exit
    help_EOF = help_exit

    def get_full_shell_command(self, cmd: str):
        return self._app.get_self_executable_script() + " " + cmd

    def _invoke_command(
        self, command_path: str, command: str, output: str = "implicit", default=None
    ):
        try:
            command_object = self._app.get_commands().get_command_by_path(
                path=command_path.split(), return_first_executable=True
            )
            if command_object is None:
                raise AppModeException(
                    "Command not be found in config: " + command_path + " " + command
                )

            if (
                "args" in command_object
                and command_object.args.config.get("verbose", True)
                and command not in ["h", "cli_info"]
            ):
                print("Command: " + command_path + " " + command)

            return self._invoke_by_command_object(command_object, command, default)
        except (AppModeException, ToolsException):
            # todo: implement verbose version print(e)
            if command in ["h", "cli_info"]:
                return default
            return self._invoke_shell(command_path + " " + command, output)
        except Exception:
            # todo: implement verbose version print(e)
            return self._invoke_shell(command_path + " " + command, output)

    def _invoke_by_command_object(self, command: Command, method: str, default=None):
        if command is None:
            raise AppModeException("Could not invoke empty command")

        command_builder_factory = self._app.get_command_builder_factory()
        command_object = command_builder_factory.create(command).build(command)

        if hasattr(command_object, method):
            method_obj = getattr(command_object, method)
            result = method_obj()
            return result

        raise AppModeException("The command was not invocable on command object")

    def _invoke_shell(self, command: str, output: str = "implicit"):
        try:
            full_command = self.get_full_shell_command(command)
            command_runner = self._create_shell_command_runner_object(
                full_command, output
            )
            return command_runner()
        except Exception as e:
            self._app.logger.warning(e)
            return 0

    def _create_shell_command_runner_object(
        self, command: str, output: str = "implicit"
    ) -> CommandRunner:
        command_data = {
            Command.Constants.TYPE: Command.Constants.TYPES.CLI,
            Command.Constants.ARGS: {
                CommandArgs.Constants.COMMAND: command,
                CommandArgs.Constants.CWD: self._app.get_working_directory(),
                CommandArgs.Constants.COMMAND_CONFIG: {"output": output},
            },
        }
        return CommandRunner(
            Command("shell", command_data),
            {},
            command_input_reader_factory=self._app.get_object_manager().provide(
                ParametersReader.Factory
            ),
        )

    def _print_command_result(self, returned_value=None):
        if returned_value is None:
            return

        if type(returned_value) is list:
            for key in returned_value:
                print(key)
            return

        if type(returned_value) is int:
            if returned_value == 0:
                print(self._app.__.green("Done"))
            else:
                print(self._app.__.red("With errors"))



from abc import abstractmethod

from lcli.config import CommandParam


class CommandInterface:
    @abstractmethod
    def do_something(self):
        pass


class CommandRunnerInterface:
    @abstractmethod
    def do_something(self):
        pass


class CommandWrapperInterface:
    pass


class CommandBuilderInterface:
    pass


class CommandBuilderFactoryInterface:
    pass


class CommandException(Exception):
    pass


class RequiredParameterException(CommandException):
    def __init__(self, parameter: CommandParam):
        self.parameter = parameter



from lcli.config import CommandParam


class CommandException(Exception):
    pass


class RequiredParameterException(CommandException):
    def __init__(self, parameter: CommandParam):
        self.parameter = parameter



import copy
import os
from typing import Dict

from lcli.exceptions import LcliException


class ConfigException(LcliException):
    pass


class ConfigNamespaces(object):
    COMMANDS: str = "commands"


class Config(object):
    """
    Base config command for projects, extend to achieve the needed flexible, extensible data object
    """

    _config: dict

    def __init__(self, config_data: dict = {}):
        self._config = config_data

    def __getitem__(self, key: str) -> "Config":
        return self._config[key]

    def __contains__(self, key: str) -> bool:
        return self._config.__contains__(key)

    def __len__(self):
        return len(self._config)

    def __iter__(self):
        return self._config.__iter__()

    def get_property(self, property_name):
        """
        Return the value of a given key
        :param property_name:
        :return:
        """
        if property_name not in self._config.keys():
            return None
        return self._config[property_name]

    @classmethod
    def _parse_path(cls, path=None):
        """
        Convert the path to list of config codes
        :param path:
        :return:
        """
        if path is None:
            path = []
        if type(path) == str:
            path = path.split("/")
        return path

    def get_all(self) -> dict:
        """

        :return: dict
        """
        return self._config

    def get(self, path=None, default=None):
        """
        Return the config found on a given path
        :param path: can be a string separated by / or list of config codes on the needed path
        :param default:
        :return: mixed
        """
        path_list = self._parse_path(path)
        config = self._config
        current_path = ""
        for key in path_list:
            current_path += key
            if config is not None and key in config:
                config = config[key]
            else:
                if default is None:
                    raise ConfigException(
                        'Requested config path not found "' + current_path + '". '
                        'The full path was: "' + path + '".'
                    )
                else:
                    return default
            current_path += "/"

        if config is None:
            return default
        return config

    def overwrite(self, other_config: "Config") -> "Config":
        config_to_merge = other_config.get()
        for k in config_to_merge:
            config_value = config_to_merge[k]
            if k in self._config and isinstance(self._config[k], dict):
                self._config[k] = dict_merge(self._config[k], config_value)
            else:
                self._config[k] = copy.deepcopy(config_value)
        return self


class CommandsConfig(Config):
    """
    Used as commands config holder on the application for commands namespace scope
    A config namespace will be a file name found and loaded form configs directories defined: in global, user, or local place
    """

    class Constants:
        COMMANDS: str = "commands"
        COMMANDS_WRAPPERS: str = "commands_wrappers"
        AVAILABLE_GROUPS: str = "available_groups"
        AVAILABLE_COMMANDS: str = "available_commands"

    def get_commands(self):
        return self.get(self.Constants.COMMANDS)

    def get_commands_wrappers(self):
        return self.get(self.Constants.COMMANDS_WRAPPERS, default={})

    def get_available_groups(self):
        return self.get(self.Constants.AVAILABLE_GROUPS, default={})

    def get_available_commands(self):
        return self.get(self.Constants.AVAILABLE_COMMANDS, default={})

    def is_command_group_available(self, group_name: str) -> bool:
        available_groups = self.get_available_groups()
        if group_name != "" and (
            len(available_groups) == 0
            or available_groups == "all"
            or "all" in available_groups
            or group_name in available_groups
        ):
            return True
        return False

    def is_command_available(self, command_name: str) -> bool:
        available_commands = self.get_available_commands()
        if (
            len(available_commands) == 0
            or available_commands == "all"
            or "all" in available_commands
            or command_name in available_commands
        ):
            return True
        return False


class ConfigReader(object):
    """
    Read the configuration from the files, this is the base funcionality
    """

    _config_file: str
    _config_dict: dict

    def __init__(self, config_path: str, required: bool = False):
        self.required = required
        self._config_file = config_path
        self._config_dict = self.read_config()

    def read_config(self) -> dict:
        """
        Implement read and convert to dict the config file
        :return:
        """
        raise ConfigException(
            "Read config method not implemented for current files type."
        )

    def get_config(self) -> Config:
        """

        :return:
        """
        return Config(self._config_dict)


class ConfigValidationError(BaseException):
    file_path: str
    message: str
    error_path: list

    def __init__(self, file_path: str, message: str, error_path=None):
        if error_path is None:
            error_path = []
        self.error_path = error_path
        self.message = message
        self.file_path = file_path


class YamlConfigReader(ConfigReader):
    _schema_file: str

    def __init__(self, config_path: str, required: bool = False, schema_file: str = ""):
        self._schema_file = schema_file
        super().__init__(config_path, required)

    def read_config(self) -> dict:
        import json

        import yaml
        from jsonschema import ValidationError, validate

        schema = {}
        if os.path.exists(self._schema_file):
            with open(self._schema_file) as f:
                schema = json.load(f)

        config = {}
        if os.path.exists(self._config_file):
            with open(self._config_file, "r") as yml_file_stream:
                config = yaml.safe_load(yml_file_stream)
            try:
                validate(config, schema)
            except ValidationError as e:
                raise ConfigValidationError(self._config_file, e.message, list(e.path))
        else:
            if self.required:
                raise ConfigException(
                    "Invalid configuration file: " + self._config_file
                )
        return config


class Command(Config):
    class Constants:
        COMMANDS: str = "commands"
        NAME: str = "name"
        DESCRIPTION: str = "description"
        TYPE: str = "type"

        class TYPES:
            CLI = "cli"
            LCLI: str = "lcli"
            WRAPPER: str = "wrapper"
            FUNCTION: str = "function"

        ARGS: str = "args"

        EXECUTABLE_COMMAND_TYPES: list = [TYPES.LCLI, TYPES.FUNCTION, TYPES.WRAPPER]

    def __init__(self, command_name: str, config_data: dict):
        config_data[self.Constants.NAME] = command_name
        args = {}
        if self.Constants.ARGS in config_data:
            args = config_data[self.Constants.ARGS]
        config_data[self.Constants.ARGS] = CommandArgs(args)
        if self.Constants.TYPE not in config_data:
            raise ConfigException(
                "Invalid config data. Command type not provided for " + command_name
            )

        if self.Constants.COMMANDS not in config_data:
            config_data[self.Constants.COMMANDS] = CommandCollection({})
        super().__init__(config_data)

    def has_commands(self):
        return len(self) > 0

    def is_executable(self):
        return self.type in self.Constants.EXECUTABLE_COMMAND_TYPES

    @property
    def type(self):
        return self.get_property(self.Constants.TYPE)

    @property
    def commands(self):
        return self.get_property(self.Constants.COMMANDS)

    @property
    def name(self):
        return self.get_property(self.Constants.NAME)

    @property
    def description(self):
        return self.get_property(self.Constants.DESCRIPTION)

    @property
    def args(self) -> "CommandArgs":
        return self.get_property(self.Constants.ARGS)

    def is_group(self):
        return self.type == "group"


class CommandArgs(Config):
    class Constants:
        PARAMS: str = "params"
        EXEC_TYPE: str = "exec_type"
        COMMAND: str = "command"
        CWD: str = "cwd"
        COMMAND_CONFIG: str = "config"

    def __init__(self, config_data: dict):
        if self.Constants.PARAMS in config_data:
            config_data[self.Constants.PARAMS] = CommandArgsParams(
                config_data[self.Constants.PARAMS]
            )
        if self.Constants.COMMAND_CONFIG in config_data:
            config_data[self.Constants.COMMAND_CONFIG] = Config(
                config_data[self.Constants.COMMAND_CONFIG]
            )
        super().__init__(config_data)

    @property
    def exec_type(self) -> str:
        return self.get_property(self.Constants.EXEC_TYPE)

    @property
    def command(self) -> str:
        return self.get_property(self.Constants.COMMAND)

    @property
    def config(self) -> Config:
        return self.get(self.Constants.COMMAND_CONFIG, Config())

    @property
    def params(self) -> "CommandArgsParams":
        return self.get_property(self.Constants.PARAMS)


class CommandArgsParams(Config):
    class Constants:
        pass

    def __init__(self, params: dict):
        _params = {}
        for key in params:
            _params[key] = CommandParam(params[key])
        super().__init__(_params)

    def get_param(self, key: str) -> "CommandParam":
        """

        :param key:
        :return: CommandConfig
        """
        return self.get(key)


class CommandParam(Config):
    class Constants:
        TYPE: str = "type"
        NAME: str = "name"
        OUTPUT_FORMAT: str = "output_format"
        MESSAGE: str = "message"
        CHOICES: str = "choices"
        CHOICES_CMD: str = "choices_cmd"

    def __init__(self, config_data: dict):
        if self.Constants.CHOICES in config_data:
            config_data[self.Constants.CHOICES] = CommandParamChoices(
                config_data[self.Constants.CHOICES]
            )
        if self.Constants.CHOICES_CMD in config_data:
            config_data[self.Constants.CHOICES_CMD] = CommandParamChoicesCmd(
                config_data[self.Constants.CHOICES_CMD]
            )
        super().__init__(config_data)

    @property
    def type(self) -> str:
        return self.get_property(self.Constants.TYPE)

    @property
    def name(self) -> str:
        return self.get_property(self.Constants.NAME)

    @property
    def output_format(self) -> str:
        return self.get_property(self.Constants.OUTPUT_FORMAT)

    @property
    def message(self) -> str:
        return self.get_property(self.Constants.MESSAGE)

    @property
    def choices(self) -> "CommandParamChoices":
        return self.get_property(self.Constants.CHOICES)

    @property
    def choices_cmd(self) -> "CommandParamChoicesCmd":
        return self.get_property(self.Constants.CHOICES_CMD)


class CommandParamChoices(Config):
    class Constants:
        pass

    def __init__(self, choices: dict):
        _params = {}
        for choice in choices:
            _params[choice[CommandParamChoice.Constants.NAME]] = CommandParamChoice(
                choice
            )
        super().__init__(_params)

    def get_choice(self, key: str) -> "CommandParamChoice":
        """

        :param key:
        :return: CommandConfig
        """
        return self.get(key)


class CommandParamChoice(Config):
    class Constants:
        NAME: str = "name"
        VALUE: str = "value"

    @property
    def name(self) -> str:
        return self.get_property(self.Constants.NAME)

    @property
    def value(self) -> str:
        return self.get_property(self.Constants.VALUE)


class CommandParamChoicesCmd(Config):
    class Constants:
        COMMAND: str = "cmd"
        FILTER: str = "filter"

    def is_empty(self):
        return len(self._config) == 0

    @property
    def command(self) -> str:
        return self.get_property(self.Constants.COMMAND)

    @property
    def filter(self) -> dict:
        return self.get_property(self.Constants.FILTER)


class CommandCollection(Config):
    _config: Dict[str, Command]

    def __init__(self, commands: dict):
        _commands = {}
        if "list_type" in commands:
            del commands["list_type"]
        for command_key in commands:
            if Command.Constants.COMMANDS in commands[command_key]:
                sub_commands = CommandCollection(
                    commands[command_key][Command.Constants.COMMANDS]
                )
                commands[command_key][Command.Constants.COMMANDS] = sub_commands

            command = commands[command_key]

            if "cli_code" in command:
                cli_code = command["cli_code"]
                if cli_code is not None and cli_code != "":
                    command_key = cli_code

            _commands[command_key] = Command(command_key, command)
        super().__init__(_commands)

    def get_all(self) -> Dict[str, Command]:
        """

        :return: Dict[str, CommandConfig]
        """
        return self._config

    def get_command(self, command_key: str) -> Command:
        """

        :param command_key:
        :return: CommandConfig
        """
        return self.get(command_key)

    def add_command(self, command: Command) -> "CommandCollection":
        """
        Append a command in current list

        :param command: CommandConfig
        :return: CommandConfig
        """
        self._config[command.name] = command
        return self

    def get_commands_names_in_path(
        self, path: list = None, ignored_commands=None, available_groups: list = []
    ) -> dict:
        """
        Get commands names available in a given path

        :param path:
        :param ignored_commands:
        :param available_groups:
        :return: dict containing commands names
        """
        if ignored_commands is None:
            ignored_commands = ["-i", "-v", "-version", "--version", "h", "cli_info"]
        if path is None:
            path = []
        try:
            commands_list = self._get_command_collection_by_path(self, path)
            return self.get_first_level_commands(
                commands_list, ignored_commands, available_groups=available_groups
            )
        except ConfigException:
            return {}

    def get_first_level_commands(
        self,
        commands_list: "CommandCollection",
        ignored_commands: list,
        available_groups: list = [],
    ) -> dict:
        """
        Returns the directly/first accessible commands for a command list
        Commands of type group are not needed to address a command in commands hierarchy
        For this reason the first level command for a command list are the first level commands that are not group

        :param commands_list:
        :param ignored_commands:
        :param available_groups:
        :return: dict containing the first level commands described above
        """
        result = {}
        if type(commands_list) == CommandCollection:
            for key in commands_list:
                if key in ignored_commands:
                    continue
                command = commands_list.get_command(key)
                if command.is_group():
                    if command.name in available_groups:
                        sub_class_commands = self.get_first_level_commands(
                            command.commands, ignored_commands
                        )
                        for sub_command_key in sub_class_commands:
                            result[sub_command_key] = sub_class_commands[
                                sub_command_key
                            ]
                else:
                    result[command.name] = command.name
        return result

    def get_command_collection_by_path(
        self, path: list = None, return_deepest=False
    ) -> "CommandCollection":
        """
        Return the list of commands accessed by a given path for invocation

        :param path: a list of strings representing the command codes
        :param return_deepest: - return the deepest found path
        :return:
        :raise: lcli.exception.ConfigException
        """
        if path is None:
            path = []
        return self._get_command_collection_by_path(
            self, path, return_deepest=return_deepest
        )

    def _get_command_collection_by_path(
        self,
        commands_list: "CommandCollection",
        path: list = None,
        return_deepest=False,
    ) -> "CommandCollection":
        """
        Return the list of commands accessed by a given path for invocation

        :param commands_list:
        :param path: a list of strings representing the command codes
        :param return_deepest: - return the deepest found path
        :return:
        :raise: lcli.exception.ConfigException
        """
        key_index = 0
        if path is None:
            path = []
        path = list(path)
        for key in path:
            if commands_list and len(commands_list) > 0 and key in commands_list:
                command = commands_list.get_command(key)
                commands_list = command.commands
            else:
                found_command = False
                for command_key in commands_list:
                    command = commands_list.get_command(command_key)
                    if command.is_group():
                        commands_list = self._get_command_collection_by_path(
                            command.commands, path[key_index:]
                        )
                        found_command = True
                        break
                if not found_command:
                    if return_deepest:
                        return commands_list
                    raise ConfigException("Broken command path")
            key_index += 1
        return commands_list

    def get_command_by_path(self, path: list = None, return_first_executable=False):
        """
        Return the command accessed by a given path

        :param path: a list of strings representing the command codes
        :param return_first_executable:
        :return:
        :raise: lcli.exception.ConfigException
        """
        if path is None:
            path = []
        return self._get_command_by_path(
            self, path, return_first_executable=return_first_executable
        )

    def _get_command_by_path(
        self,
        commands_list: "CommandCollection",
        path: list = None,
        return_first_executable=False,
    ):
        """
        Return the command accessed by a given path

        :param commands_list:
        :param path: a list of strings representing the command codes
        :param return_first_executable: - return the deepest found path
        :return:
        :raise: lcli.exception.ConfigException
        """
        if path is None:
            path = []
        path = list(path)
        path_len = len(path)
        if path_len == 0:
            return None
        current_path = path[0]
        command = None
        if commands_list and len(commands_list) > 0 and current_path in commands_list:
            command = commands_list.get_command(current_path)
            if return_first_executable and command.is_executable():
                return command
            if path_len > 1:
                command = self._get_command_by_path(
                    command.commands,
                    path[1:],
                    return_first_executable,
                )
        else:
            for command_key in commands_list:
                command_i = commands_list.get_command(command_key)
                if command_i.is_group():
                    command = self._get_command_by_path(
                        command_i.commands,
                        path,
                        return_first_executable,
                    )
                    if command is not None:
                        return command

        if command is not None:
            return command

        raise ConfigException("Could not identify the command by path")


class CommandWrapper(Config):
    class Constants:
        HANDLER: str = "handler"
        BUILDER: str = "builder"
        ARGS_CONFIG: str = "args_config"

    @property
    def handler(self) -> str:
        return self.get_property(self.Constants.HANDLER)

    @property
    def args_config(self) -> str:
        return self.get_property(self.Constants.ARGS_CONFIG)

    @property
    def builder(self) -> str:
        return self.get_property(self.Constants.BUILDER)


class CommandWrapperCollection(Config):
    class Constants:
        WRAPPERS = "wrappers"

    def __init__(self, wrappers: dict):
        _wrappers = {}
        for key in wrappers:
            _wrappers[key] = CommandWrapper(wrappers[key])
        super().__init__(_wrappers)


def dict_merge(a, b):
    """recursively merges dict's. not just simple a['key'] = b['key'], if
    both a and behave a key who's value is a dict then dict_merge is called
    on both values and the result stored in the returned dictionary."""
    if not isinstance(b, dict):
        return b
    result = copy.deepcopy(a)
    for k, v in b.items():
        if k in result and isinstance(result[k], dict):
            result[k] = dict_merge(result[k], v)
            continue

        # todo: implement list merging
        # if isinstance(v, list) and k in result and isinstance(result[k], list):
        #     result[k] = dict_list(result[k], v)
        #     continue

        result[k] = copy.deepcopy(v)
    return result


def dict_list(list1: list, list2: list):
    result = {}
    for value1 in list1:
        result[value1] = value1
    for value2 in list2:
        result[value2] = value2

    return list(result.values())



import os
import sys
import traceback

from .__init__ import EXECUTABLE_NAME


def main(init_params: dict = None):
    cwd = process_working_directory()
    if init_params is None:
        init_params = {}
    try:
        from lcli.app import App

        project_root_path = os.path.realpath(os.path.dirname(__file__))
        app = App(
            app_path=project_root_path,
            executable_name=EXECUTABLE_NAME,
            init_params=init_params,
            cwd=cwd,
        )
        app.logger.info("start")
        app.run()
        app.logger.info("end")
    except ModuleNotFoundError as error:
        print("\n")
        print("Module dependencies are not met: " + str(error))
        print(
            "Please install all requirements defined in requirements.txt of "
            + EXECUTABLE_NAME
            + " before running the application."
        )
        print("\n")

    except Exception as error:
        print("Unexpected error: " + str(error))
        print(traceback.format_exc())
    except KeyboardInterrupt:
        print("\n")
        print("Execution stopped by keyboard interrupt signal.")
        print("\n")


def process_working_directory():
    cwd = os.getcwd()

    found_cwd = False
    to_remove_args = []

    for param in sys.argv:
        if found_cwd:
            to_remove_args.append(param)
            cwd = param
            break
        if param == "--cwd":
            to_remove_args.append(param)
            found_cwd = True

    for flag in to_remove_args:
        sys.argv.remove(flag)

    sys.path.append(cwd)
    return cwd


if __name__ == "__main__":
    main({})



import subprocess

from lcli.command.exceptions import RequiredParameterException
from lcli.config import CommandArgsParams, Config, ConfigException
from lcli.input.prompt import AskQuestions


class ParametersReader(object):
    _params: CommandArgsParams
    _defaults: Config

    class Factory(object):
        @classmethod
        def create(
            cls, params: CommandArgsParams, defaults: Config
        ) -> "ParametersReader":
            return ParametersReader(params, defaults)

    def __init__(self, params: CommandArgsParams, defaults: Config):
        self._defaults = defaults
        self._params = params

    def read(self):

        if self._params is not None:
            questions = {}
            answers = {}
            for param in self._params:
                param_object = self._params.get_param(param)

                if param_object.type == "config":
                    try:
                        answers[param] = self._defaults.get(
                            param_object.get("config_path")
                        )
                    except ConfigException:
                        if param_object.get("required", False):
                            raise RequiredParameterException(param_object)
                        continue

                if param_object.type == "input":
                    questions[param] = {
                        "type": "input",
                        "name": param,
                        "message": param_object.message,
                    }

                if param_object.type == "list" or param_object.type == "autocomplete":
                    question = param_object.get_all().copy()
                    question["name"] = param
                    if (
                        param_object.choices_cmd is not None
                        and not param_object.choices_cmd.is_empty()
                    ):
                        choices_cmd = param_object.choices_cmd
                        stdout = subprocess.check_output(choices_cmd.command.split())
                        out = stdout.decode()
                        choices_result = [
                            b.strip(choices_cmd.filter["strip"])
                            for b in out.splitlines()
                        ]
                        question["choices"] = choices_result
                    else:
                        question["choices"] = {}
                        if param_object.choices is not None:
                            for k in param_object.choices.get_all():
                                choice = param_object.choices.get_choice(k)
                                question["choices"][choice.value] = choice.name
                    question["answers"] = {}
                    question["type"] = "autocomplete"
                    questions[param] = question

            questions_object = AskQuestions(questions)
            qa = questions_object.ask()
            for q in qa:
                answers[q] = qa[q]
            return answers

        return {}



class InputException(Exception):
    pass


class InterruptedInputException(InputException):
    def __init__(self, message: str, partial_answers: dict, errors=None):
        if errors is None:
            errors = []
        self.errors = errors
        self.partial_answers = partial_answers
        self.message = message
        super().__init__(self.message)



from pydoc import locate
from typing import Dict, List

from lcli.app import App
from lcli.config import Command
from lcli.exceptions import BuilderException


class BaseBuilder:
    """
    Build various type of objects to use them in Fire application mode
    Mandatory to define the command_type for each builder
    """

    _app: App
    command_type: str = "n/a"

    def __init__(self, app: App) -> None:
        self._app = app

    def can_build(self, command: Command) -> bool:
        return command.type == self.command_type

    def build(self, command: Command):
        raise BuilderException("Command builder not implemented for: " + command.name)

    @classmethod
    def locate(cls, path: str):
        try:
            return locate(path)
        except Exception:
            raise BuilderException(
                "Invalid configuration provided for command. Path "
                + path
                + " could not be loaded."
            )


class CommandBuilderFactory(object):
    """
    Create instances of builders responsible to construct objects for fire application mode
    """

    _app: App
    _builders_available: List[BaseBuilder] = []
    _builders_instances_cache: Dict[str, BaseBuilder] = {}

    def __init__(self, app: App) -> None:
        self._app = app

    @staticmethod
    def register_builder(class_ref):
        """
        Register your custom builder for command type, this should be called before first usage

        :param class_ref:
        :return:
        """
        CommandBuilderFactory._builders_available.append(class_ref)
        return CommandBuilderFactory

    def create(self, command: Command) -> BaseBuilder:
        """
        Create the first builder that can construct the command object for current command

        :param command:
        :return:
        """
        for builder in self._builders_available:
            builder_object = self.get_builder(builder)
            if builder_object.can_build(command):
                return builder_object
        raise BuilderException(
            "Unable to find a proper command builder for command: " + command.name
        )

    def get_builder(self, builder) -> BaseBuilder:
        """
        Return builder object by class, and keep cache of instantiated objects

        :param builder:
        :return:
        """
        if builder.command_type not in self._builders_instances_cache:
            self._builders_instances_cache[builder.command_type] = builder(self._app)
        return self._builders_instances_cache[builder.command_type]


class LcliBuilder(BaseBuilder):
    """
    This builder is responsible to construct objects located in lcli.tools namespace
    Usually extended form lcli.tools.Base.BaseTool
    """

    command_type: str = Command.Constants.TYPES.LCLI

    def build(self, command: Command):
        located_object = locate(command.args.command)
        if callable(located_object):
            return located_object(self._app)


class SimpleBuilder(BaseBuilder):
    """
    A simple command object, without other dependencies
    """

    command_type: str = Command.Constants.TYPES.FUNCTION

    def canBuild(self, command: Command) -> bool:
        if command.type is None:
            return True
        return super().can_build(command)

    def build(self, command: Command):
        return locate(command.args.command)


class WrappersBuilder(BaseBuilder):
    """
    This will build a "virtual" command object capable to execute commands form configuration
    """

    command_type: str = Command.Constants.TYPES.WRAPPER

    def build(self, command: Command):
        try:
            wrappers = self._app.get_commands_wrappers()
            wrapper_object = wrappers.get(command.args.get("wrapper"))
            builder_name = wrapper_object.get("builder")
            if (
                builder_name is not None
                and type(builder_name) == str
                and len(builder_name)
            ):
                builder = locate(builder_name)
                if callable(builder):
                    builder_object = builder(self._app)
                    if callable(builder_object):
                        return builder_object.build(command)

            handler = locate(wrapper_object.get("handler"))
            if callable(handler):
                return handler(self._app, command)

        except Exception as e:
            self._app.logger.warning(e)
            raise BuilderException("Unable to build command: " + command.name)



class Calculator(object):
    """A simple calculator class."""

    def double(self, number):
        """Double the input -> twice"""
        return 2 * number

    def h(self):
        import json

        print(json.dumps(["double", "Calculator12"], sort_keys=True))
        exit(0)

    class Calculator2(object):
        """A simple calculator class."""

        def double(self, number):
            return 2 * number



from lcli.command.builders import (
    CommandBuilderFactory,
    LcliBuilder,
    SimpleBuilder,
    WrappersBuilder,
)


CommandBuilderFactory.register_builder(LcliBuilder)
CommandBuilderFactory.register_builder(SimpleBuilder)
CommandBuilderFactory.register_builder(WrappersBuilder)



import json
import subprocess
from string import Template

import pinject

from lcli.command.exceptions import RequiredParameterException
from lcli.command.input import ParametersReader
from lcli.config import Command as ConfigCommand
from lcli.config import Config
from lcli.input.exctptions import InterruptedInputException


class CommandRunner:
    """
    Run a shell command using application configuration, and dynamic parameters reader
    """

    _command_input_reader_factory: ParametersReader.Factory
    _command: ConfigCommand
    _local_variables: Config

    @pinject.inject(all_except=["command", "local_variables"])
    def __init__(
        self,
        command: ConfigCommand,
        local_variables: dict,
        command_input_reader_factory: ParametersReader.Factory,
        parent_command: ConfigCommand = None,
    ):
        self._command_input_reader_factory = command_input_reader_factory
        self._command = command
        self._parent_command = parent_command
        self._local_variables = Config(local_variables)

    def __call__(self, **args: str):
        try:
            command_input_reader = self._command_input_reader_factory.create(
                self._command.args.params, self._local_variables
            )
            answers = command_input_reader.read()
            command = ""
            if (
                self._parent_command is not None
                and self._parent_command.args is not None
            ):
                command += self._parent_command.args.get("commands_prefix", "")
            command += " " + self._command.args.command

            command = self.append_parameters_to_command(command, answers, args)
            output_type = self._command.args.config.get("output", "implicit")
            if "args" in args:
                command += " " + args["args"]

            cwd = self._command.args.get("cwd", ".")
            # print("Command [" + cwd + "]: " + command)

            if self._command.type == "bash-script":
                script = self._command.args.get("script")
                process = subprocess.Popen(script, shell=True, cwd=cwd)
                process.communicate()
                return

            if self._command.name == "cli_info":
                output_type = "list"

            if output_type is None or output_type == "implicit":
                process = subprocess.Popen(command, shell=True, cwd=cwd)
                process.communicate()
                return

            if output_type == "list":
                process = subprocess.Popen(
                    command,
                    shell=True,
                    cwd=cwd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.DEVNULL,
                )
                out, err = process.communicate()
                if process.returncode != 0:
                    return []

                processed_out = str(out.rstrip().decode())
                list_parts = [
                    y for y in (x.strip() for x in processed_out.splitlines()) if y
                ]
                return list_parts

            if output_type == "json":
                p = subprocess.Popen(
                    command,
                    shell=True,
                    cwd=cwd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.DEVNULL,
                )
                out, err = p.communicate()
                if p.returncode == 0:
                    return json.loads(str(out.rstrip().decode()))
        except InterruptedInputException:
            return []
        except RequiredParameterException as e:
            print("Parameter " + e.parameter.name + "is required")
            return []

    def append_parameters_to_command(self, command: str, answers: dict, args: dict):
        params = self._command.args.params
        if params is not None:
            for param in params:
                param_object = params.get_param(param)
                param_name = param_object.name
                if param_name != "":
                    if param_name[1:] in args:
                        param_value = args[param_name[1:]]
                        if param_object.output_format is not None and len(
                            param_object.output_format
                        ):
                            value_template = Template(param_object.output_format)
                            command = command + value_template.substitute(
                                {"value": param_value, "name": param_name}
                            )
                        else:
                            command = (
                                command + " " + param_name + ' "' + param_value + '"'
                            )
                        continue

                answer_list = [command]

                answer_output = ""
                if param_object.output_format is not None and len(
                    param_object.output_format
                ):
                    if param in answers:
                        value_template = Template(param_object.output_format)
                        answer_output = value_template.substitute(
                            {"value": answers[param], "name": param_name}
                        )
                else:
                    if param in answers:
                        answer_output = answers[param]

                answer_list.append(answer_output)

                command = " ".join(answer_list)

        command_template = Template(command)
        processed_local_variables = {}
        local_variables = self._local_variables.get_all()
        for key in local_variables:
            if type(local_variables[key]) in [dict]:
                string_param = ""
                for param_key in local_variables[key]:
                    for variable_value in local_variables[key][param_key]:
                        if len(string_param):
                            string_param += " "
                        string_param += param_key + " " + variable_value
                processed_local_variables[key] = string_param
                continue

            if type(local_variables[key]) in [list]:
                string_param = ""
                for index in local_variables[key]:
                    string_param += " " + local_variables[key][index]
                processed_local_variables[key] = string_param
                continue

            processed_local_variables[key] = local_variables[key]

        command = command_template.substitute(**processed_local_variables)

        return command

    @classmethod
    def parse_args(cls, arg_string: str):
        import argparse

        parser = argparse.ArgumentParser(description="Argparse Test script")
        parser.add_argument("-json", action="store_true")

        args = parser.parse_args(arg_string.split())
        return args



from typing import Any, List, Optional, Tuple, TypeVar

from prompt_toolkit import Application
from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion import Completer
from prompt_toolkit.filters import FilterOrBool
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.key_binding import KeyBindings, merge_key_bindings
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous
from prompt_toolkit.key_binding.defaults import load_key_bindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import D, HSplit, Layout
from prompt_toolkit.layout.containers import AnyContainer
from prompt_toolkit.shortcuts.dialogs import _return_none
from prompt_toolkit.styles import BaseStyle
from prompt_toolkit.widgets import Button, Dialog, Label, RadioList, TextArea


_T = TypeVar("_T")


class LcliDialog(RadioList):
    def _handle_enter(self):
        super()._handle_enter()
        get_app().exit(result=self.current_value)


class RadioListButtonHandler(object):
    def __init__(self, button: dict, radio_list: RadioList):
        self.radio_list = radio_list
        self.button = button

    def __call__(self, *args, **kwargs) -> None:
        get_app().exit(result=self.button["text"])


def radiolist_dialog(
    title: AnyFormattedText = "",
    text: AnyFormattedText = "",
    values: Optional[List[Tuple[_T, AnyFormattedText]]] = None,
    buttons=None,
    style: Optional[BaseStyle] = None,
) -> Application[_T]:
    """
    Display a simple list of element the user can choose amongst.

    Only one element can be selected at a time using Arrow keys and Enter.
    The focus can be moved between the list and the Ok/Cancel button with tab.
    """
    if buttons is None:
        buttons = {}
    if values is None:
        values = []

    radio_list = LcliDialog(values)

    buttons_list = []
    for key in buttons:
        button = buttons[key]
        handler = RadioListButtonHandler(button, radio_list)
        if "handler" in button:
            handler = button["handler"](button, radio_list)
        button_object = Button(text=button["text"], handler=handler)
        buttons_list.append(button_object)

    dialog = Dialog(
        title=title,
        body=HSplit([Label(text=text, dont_extend_height=True), radio_list], padding=1),
        buttons=buttons_list,
        with_background=True,
    )

    return _create_app(dialog, style)


def input_dialog(
    title: AnyFormattedText = "",
    text: AnyFormattedText = "",
    ok_text: str = "OK",
    cancel_text: str = "Cancel",
    completer: Optional[Completer] = None,
    password: FilterOrBool = False,
    style: Optional[BaseStyle] = None,
) -> Application[str]:
    """
    Display a text input box.
    Return the given text, or None when cancelled.
    """

    def accept(buf: Buffer) -> bool:
        get_app().layout.focus(ok_button)
        return True  # Keep text.

    def ok_handler() -> None:
        get_app().exit(result=textfield.text)

    ok_button = Button(text=ok_text, handler=ok_handler)
    cancel_button = Button(text=cancel_text, handler=_return_none)

    textfield = TextArea(
        multiline=False, password=password, completer=completer, accept_handler=accept
    )

    dialog = Dialog(
        title=title,
        body=HSplit(
            [
                Label(text=text, dont_extend_height=True),
                textfield,
            ],
            padding=D(preferred=1, max=1),
        ),
        buttons=[ok_button, cancel_button],
        with_background=True,
    )

    return _create_app(dialog, style)


def _create_app(dialog: AnyContainer, style: Optional[BaseStyle]) -> Application[Any]:
    # Key bindings.
    bindings = KeyBindings()
    bindings.add("tab")(focus_next)
    bindings.add("s-tab")(focus_previous)

    @bindings.add(Keys.Escape)
    @bindings.add("c-c")
    @bindings.add("c-q")
    def _(event):
        """Pressing Ctrl-Q or Ctrl-C will exit the user interface."""
        event.app.exit()

    return Application(
        layout=Layout(dialog),
        key_bindings=merge_key_bindings(
            [
                load_key_bindings(),
                bindings,
            ]
        ),
        mouse_support=True,
        style=style,
        full_screen=True,
    )



import datetime
import errno
import hashlib
import os
import shelve

from lcli.exceptions import LcliException


class CacheNotFoundException(LcliException):
    pass


class Cache(object):
    lifetime: int
    cache_key: str

    class Constants:
        EXPIRE_ON = "expire_on"
        LIFETIME = "lifetime"
        DATA = "data"

    def __init__(self, cache_directory: str, cache_keys: list, lifetime: int = 0):
        """

        :type cache_keys: list
        :type cache_directory: str
        """
        self.lifetime = lifetime
        self.cache_key = self.generate_cache_key(cache_keys)
        try:
            os.makedirs(cache_directory)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        self.cache_path = os.path.join(cache_directory, self.cache_key)

    def save(self, data):
        cache = shelve.open(self.cache_path)
        cache[self.cache_key] = {
            self.Constants.DATA: data,
            self.Constants.LIFETIME: self.lifetime,
            self.Constants.EXPIRE_ON: datetime.datetime.now()
            + datetime.timedelta(self.lifetime),
        }
        cache.close()

    @classmethod
    def generate_cache_key(cls, cache_keys: list) -> str:
        md5 = hashlib.md5(usedforsecurity=False)

        if len(cache_keys) == 1:
            md5.update(cache_keys[0].encode("utf-8"))
        else:
            md5.update("".join(cache_keys).encode("utf-8"))
        return md5.hexdigest()

    def get(self):
        result = None
        cache = shelve.open(self.cache_path)
        if self.cache_key in cache:
            cache_data = cache[self.cache_key]

            if (
                cache_data[self.Constants.LIFETIME] == 0
                or cache_data[self.Constants.EXPIRE_ON] > datetime.datetime.now()
            ):
                result = cache_data[self.Constants.DATA]
            else:
                del cache[self.cache_key]

        cache.close()

        if result is not None:
            return result

        raise CacheNotFoundException("Cache not found")



from lcli.app import App
from lcli.command.builders import BaseBuilder
from lcli.command.input import ParametersReader
from lcli.command.subprocess import CommandRunner
from lcli.config import Command
from lcli.tools.base import BaseTool


class BaseCommandWrapperInterface(object):
    pass


class BashCommandWrapperInterface(BaseCommandWrapperInterface):
    pass


class BashCommandWrapper(BaseTool, BashCommandWrapperInterface):
    """
    Run command from configuration, commands with type wrapper
    This will consist of executing shell commands on a given directory using configuration, parameters completion and
    interactive selections of parameters
    """

    _command_name: str
    _command: Command

    def __init__(self, app: App, command: Command) -> None:
        super().__init__(app)
        self._command_name = command.name
        self._command = command
        self.__init_command()

    def __init_command(self):
        self.__doc__ = self._command.description

        sub_commands = self._command.commands
        for command_code in sub_commands:
            current_command = sub_commands[command_code]
            current_command_obj = CommandRunner(
                current_command,
                self._app.get_config_object().get(
                    ["commands_defaults", self._command_name], default={}
                ),
                self._app.get_object_manager().provide(ParametersReader.Factory),
                self._command,
            )
            current_command_obj.__doc__ = (
                current_command.description + " (" + current_command.args.command + ")"
            )
            setattr(self, command_code, current_command_obj)

    class _Builder(BaseBuilder):
        """
        This will be a custom builder for current wrapper, custom ones may be implemented
        """

        def build(self, command: Command) -> "BashCommandWrapper":
            return BashCommandWrapper(self._app, command)


class ManualWrapper:
    def __init__(self):
        self._dict = {}

    def __iter__(self):
        return self._dict.__iter__()

    def __setitem__(self, name, value):
        setattr(self, name, value)
        self._dict[name] = value

    def __getitem__(self, name):
        return self._dict[name]

    def __contains__(self, key):
        return key in self._dict

    def __update__(self, value):
        self._dict = value

    def __len__(self):
        return len(self._dict)



import logging


def setup_logger(name: str, verbosity: int, file: str) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(verbosity)
    handler = logging.FileHandler(file)
    handler.setLevel(verbosity)
    handler.setFormatter(
        logging.Formatter(
            "%(asctime)s " "%(levelname)-8s " "%(name)s: %(message)s",
            "%Y-%m-%d %H:%M:%S",
        )
    )
    logger.addHandler(handler)
    return logger



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['lcli',
 'lcli.api',
 'lcli.api.command',
 'lcli.app_mode',
 'lcli.command',
 'lcli.input',
 'lcli.tools']

package_data = \
{'': ['*'],
 'lcli': ['config/*',
          'config/samples/*',
          'config/samples/commands/*',
          'config/samples/prototypes/*',
          'config/schema/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'blessings>=1.7,<2.0',
 'fire>=0.4,<0.5',
 'jsonschema>=4.5,<5.0',
 'pinject>=0.14,<0.15',
 'prompt-toolkit>=3.0,<4.0',
 'pyfiglet>=0.8,<0.9',
 'types-PyYAML']

entry_points = \
{'console_scripts': ['flex-cli = flex-cli.__main__:main',
                     'lcli = lcli.__main__:main']}

setup_kwargs = {
    'name': 'flex-cli',
    'version': '0.1.2',
    'description': "Local CLI is an extendable interactive command line tool built in mind with the objective to make the interaction with day to day scripts more user friendly. The scripts you use daily should be easy to maintain, develop and document so you don't need to worry to write long documents on how to use them.",
    'long_description': '# Introduction\n\n[![PyPI version](https://img.shields.io/pypi/v/lcli.svg)](https://pypi.org/project/lcli)\n[![Build Status](https://github.com/a42ss/lcli/actions/workflows/python-package.yml/badge.svg)](https://github.com/a42ss/lcli/actions/workflows/python-package.yml)\n[![Coverage badge](https://raw.githubusercontent.com/a42ss/lcli/python-coverage-comment-action-data/badge.svg)](https://github.com/a42ss/lcli/tree/python-coverage-comment-action-data)\n[![CodeQL](https://github.com/a42ss/lcli/actions/workflows/codeql-analysis.yml/badge.svg?branch=main)](https://github.com/a42ss/lcli/actions/workflows/codeql-analysis.yml)\n[![CII Best Practices](https://bestpractices.coreinfrastructure.org/projects/6480/badge)](https://bestpractices.coreinfrastructure.org/projects/6480)\n[![Docs badge](https://img.shields.io/badge/docs-latest-brightgreen.svg)](https://lcli-docs.readthedocs.io/en/latest/)\n[![Chat badge](https://img.shields.io/badge/chat-IRC-brightgreen.svg)](https://github.com/a42ss/lcli/discussions)\n[![License](https://img.shields.io/badge/license-MIT-brightgreen.svg)](https://lcli-docs.readthedocs.io/en/latest/)\n\n   \n\nLocal development productivity tools meant to smooth and ease developers day to day work. \nIt is meant to be cross-platform but for now is tested using MacOs and Ubuntu\n\n## This are few examples\n\n* unified command line interfaces for multiple cli tools\n* interactive command line mode\n\n  - this act as an interactive wrapper on top of existing cli tools\n  - also allow extending the existing cli tools with auto-completion, input enhanced wizard or description\n  - allow switching between command namespaces in the same terminal\n\n* Implement custom CLI tools using Python, fully integrate with all LCLI tool features:\n  - just write some code class or function and configure them to be wired in application\n\n* use Fire to auto document Python objects, so all you should do is to focus on business logic\n   \n# Features\n\n## Fire mode\n\nIn fire mode the application allow user to configure a hierarchy of objects that fit its needs for various project.\nThe user can navigate and execute the hierarchy using Fire library by Google. \n"Python Fire is a library for automatically generating command line interfaces (CLIs) from absolutely any Python object."\nhttps://github.com/google/python-fire\n\n## Interactive mode\n\nMost often during the development process developers are using various tools for various projects.\nThe interactive mode purpose is to improve productivity by offering auto-completion details on the spot.\nThis is implemented on top of python cmd package: https://docs.python.org/3/library/cmd.html\n\n# Getting Started\n\n## Installation process\n\nInstall from source\n\n```bash\n# Using invoke\ninvoke install\n#from project root directory\n./install\n#or \npython3 -m pip install -r requirements.txt\npython3 -m pip install . --user\n```\n\n## Software dependencies\n\nThis is a Python package available as MIT License and is depending on following packages:\n \n* fire https://github.com/google/python-fire/releases\n* pinject https://github.com/google/pinject/releases\n* PyYAML\n* prompt_toolkit\n* pyfiglet\n* blessings\n* tk\n* appJar\n* jsonschema\n\n## Latest releases\n\n- V-0.2.2 - First released version. \n\n## API references\n\n# Build and Test\n\n## Build\n\n```bash\n# Using invoke\ninvoke build\n```\n\n## Test\n\n```bash\n# Using invoke\ninvoke test\ninvoke coverage\n# Using pytest\npy.test\npytest --cov=src/lcli/ .\n```\n\n# Use cases\n\n[<img src="https://img.youtube.com/vi/L9orYXE1nlU/hqdefault.jpg" width="50%">](https://youtu.be/L9orYXE1nlU)\n\n# Usage\n\n\n## Configuration\n\n# Author\n\n[George Babarus](https://github.com/georgebabarus)\n\n# Contribute\n\nFeel free to contribute to this project and make developer life essayer:\n- by submitting new ideas as a github issue [here](https://github.com/georgebabarus/lcli/issues/new)\n- by making pull request with specific bug fixes\n- for new features or architectural change please contact [George Babarus](https://github.com/georgebabarus) to avoid double work on any way.\n\n# Useful links\n\n* https://mypy.readthedocs.io/en/latest/generics.html#generics\n* https://code-maven.com/interactive-shell-with-cmd-in-python\n* https://www.journaldev.com/16140/python-system-command-os-subprocess-call\n* https://stackoverflow.com/questions/3262569/validating-a-yaml-document-in-python\n* https://github.com/oclif/oclif#-cli-types\n* https://medium.com/the-z/getting-started-with-oclif-by-creating-a-todo-cli-app-b3a2649adbcf\n* https://opensource.com/article/17/5/4-practical-python-libraries\n\n',
    'author': 'George Babarus',
    'author_email': 'george.babarus@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/a42ss/lcli',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import subprocess

from lcli.app import App
from lcli.tools import BaseTool


def _shell_execute(command: str, output: bool = False):
    if output:
        return subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
        )
    else:
        return subprocess.call(command, shell=True)


class Git(BaseTool):
    """Git vcs interaction"""

    _command_name = "git"

    def __init__(self, app: App) -> None:
        super().__init__(app)
        self._command_variables = self._app.get_config_object().get(
            ["commands_defaults", self._command_name], default={}
        )

    def pull(self):
        """Fetch changes from repo"""
        cmd = "git pull origin " + self.branch()
        _shell_execute(cmd)

    def fetch(self):
        """Fetch changes from repo"""
        _shell_execute("git fetch origin ")

    def commit(self, message: str = ""):
        """Commit changes with message or ask for it"""
        if message != "":
            cmd = 'git commit -m "' + message + '"'
        else:
            cmd = "git commit"
        _shell_execute(cmd)

    def add(self):
        """Add all changes to commit"""
        _shell_execute("git add .")

    def status(self):
        """Show changed files"""
        _shell_execute("git status")

    def diff(self):
        """Show changes"""
        _shell_execute("git diff")

    def save(self, message: str = ""):
        """Auto commit and push all changes"""
        self.add()
        if message != "":
            self.commit(message)
        else:
            self.commit(
                self.branch() + " " + self._command_variables.get("autocommit_message")
            )
        self.push()

    def amend(self):
        """Amend current changes"""
        self.add()
        _shell_execute("git commit --amend --reset-author")

    def cp(self, message: str = ""):
        """Auto commit and push all changes"""
        self.add()
        self.commit(message)
        self.push()

    def push(self, force: bool = False, remote: str = ""):
        """Push changes to remote repo"""
        cmd = "git push origin " + self.branch()
        if force:
            cmd += " -f"
        _shell_execute(cmd)

    def branch(self):
        """Push changes to remote repo"""
        cmd = "git branch 2> /dev/null | sed -e '/^[^*]/d' -e 's/* (.*)/\\1/'"
        process = _shell_execute(cmd, True)
        out, err = process.communicate()
        if process.returncode == 0:
            return str(out.rstrip().decode())
        return ""

    def pwd(self):
        process = _shell_execute("pwd", True)
        out, err = process.communicate()
        if process.returncode == 0:
            return str(out.strip().decode())
        return ""

    def cli_info(self):
        """Provide cli headline information"""
        return [self.pwd(), self.branch()]



from lcli.app import App


class BaseTool:
    _app: App

    def __init__(self, app: App) -> None:
        self._app = app

    def h(self):
        """return the list of callable methods for autocomplete purpose"""
        method_names = [
            attr
            for attr in dir(self)
            if (callable(getattr(self, attr)) and attr[0] != "_" and attr != "h")
        ]
        return method_names


class ToolsException(Exception):
    pass



import sys

import fire

from lcli.app_mode.base import AppModeBase, AppModeException
from lcli.command.wrappers import ManualWrapper
from lcli.config import CommandCollection, ConfigException
from lcli.exceptions import BuilderException


class FireException(AppModeException):
    pass


class Fire(AppModeBase):
    from lcli.app import App

    def __init__(self, app: App):
        AppModeBase.__init__(self, app)
        self._command_alias_performed = False

    def run(self) -> None:
        """
        Run application in fire mode, that means it is not interactive
        See the specs of fire project
        :return:
        """
        args = self._app.get_args()
        command_path_index = 0
        for part in args:
            if part == "-":
                break
            command_path_index += 1
        remaining_args = args[command_path_index + 1 :]
        path = args[1:command_path_index]
        new_args = args[:command_path_index]
        if len(remaining_args):
            new_args.append('--args="' + " ".join(remaining_args) + '"')

        if len(new_args) >= 2:
            if self._app.get_app_code() == new_args[1]:
                temp_args = new_args[:1]
                if len(new_args) >= 3:
                    temp_args += new_args[2:]
                new_args = temp_args
                self._command_alias_performed = True

        sys.argv.clear()
        for arg in new_args:
            sys.argv.append(arg)

        self.fire_expose(path)

    def fire_expose(self, path: list) -> None:
        """
        Expose to fire a limited number of commands from provided config
        Depending on selected path from commands hierarchy

        :param path:
        :return:
        """

        executable_name = ""
        if len(path) > 0:
            executable_name = path[:1].pop()
        if executable_name == "--" or executable_name == "":
            executable_name = self._app.get_app_code()

        result_commands = self.get_available_commands(path)
        if self._command_alias_performed:
            if (
                len(result_commands) == 1
                and (
                    type(result_commands) is dict
                    or type(result_commands) is ManualWrapper
                )
                and executable_name in result_commands
            ):
                if (
                    type(result_commands[executable_name]) is dict
                    or type(result_commands[executable_name]) is ManualWrapper
                ):
                    temp_result = ManualWrapper()
                    temp_result.__doc__ = self._app.get_app_description()

                    for key in result_commands[executable_name]:
                        temp_result[key] = result_commands[executable_name][key]
                    result_commands = temp_result

        fire.Fire(result_commands, None, executable_name)

    def get_available_commands(self, path: list) -> ManualWrapper:
        """
        Return the available commands in a given path in order to load just a potion of configs in Fire mode
        :param path:
        :return:
        """
        try:
            if len(path) > 0:
                command = path[:1].pop()
                if command == "--":
                    path = []

            if not len(path):
                command_collection = self._app.get_commands()
            else:
                try:
                    command_collection = CommandCollection({})
                    command_found = self._app.get_commands().get_command_by_path(
                        path=path[:1], return_first_executable=True
                    )
                    command_collection.add_command(command_found)
                except Exception:
                    command_collection = self._app.get_commands()
        except ConfigException as e:
            command_collection = CommandCollection({})
            self._app.logger.error(e)

        return self.build_commands_object_recursively(command_collection)

    def build_commands_object_recursively(
        self, commands_list: CommandCollection
    ) -> ManualWrapper:
        """
        Build command objects or references prepared to Fire from a CommandCollection object recursively
        It is needed because for now it rely on Fire documentation and the object should be prepared and ready to execute

        :param commands_list:
        :return: dict containing instances of objects in a hierarchy manner
        """
        config_object = self._app.get_config_object()

        command_builder_factory = self._app.get_command_builder_factory()
        result_commands = ManualWrapper()
        for key in commands_list:
            try:
                command = commands_list.get_command(key)
                if command.is_group():
                    if not config_object.is_command_group_available(key):
                        continue
                    group_objects = self.build_commands_object_recursively(
                        command.commands
                    )
                    if type(group_objects) is ManualWrapper:
                        group_objects.__doc__ = command.description

                    if key == "groups":
                        result_commands.__update__(group_objects)
                    else:
                        result_commands[key] = group_objects
                else:
                    if not config_object.is_command_available(key):
                        continue
                    result_commands[key] = command_builder_factory.create(
                        command
                    ).build(command)
            except BuilderException as e:
                self._app.logger.warning(e)

        return result_commands



import copy
import errno
import logging
import os
import sys

import pinject
from blessings import Terminal as Translator

from lcli.cache import Cache, CacheNotFoundException
from lcli.config import (
    CommandCollection,
    CommandsConfig,
    CommandWrapperCollection,
    Config,
    ConfigNamespaces,
    ConfigValidationError,
    YamlConfigReader,
)
from lcli.utils import setup_logger


class AppException(Exception):
    pass


class Output(object):
    @staticmethod
    def print(message: str, args: list):
        print(message.format(*args))

    @staticmethod
    def format(message: str, args: list):
        return message.format(*args)


class App:
    CACHE_LIFETIME = 86400
    _app_path: str
    _current_command: list[str] = []

    _object_manager: pinject.object_graph.ObjectGraph
    translator = __ = Translator()
    output = Output()
    logger: logging.Logger

    FLAG_INTERACTIVE = "-i"
    FLAG_VERBOSE = "-v"
    FLAG_DEBUG = "-d"

    _flags: dict[str, str] = {}
    _available_flags = [FLAG_INTERACTIVE, FLAG_VERBOSE, FLAG_DEBUG]

    _working_directory: str = ""
    _user_home_directory: str = ""
    _user_home_directory_path: str = ""

    CONF_EXECUTABLE_PATH: str = "cli_executable_path"
    CONF_EXECUTABLE_NAME: str = "cli_executable_name"
    _executable_path: str
    _executable_name: str

    _config_folder_defaults = "config"
    _config_folder_working_directory = ".lcli"
    _config_folder_defaults_path = ""
    _working_directory_conf_path: str = ""
    _config_folder_home_directory = os.path.join(".config", ".lcli")
    _config_namespace = "commands"
    _log_file = "lcli.log"
    _cache_directory_path: str

    _commands: CommandCollection
    _commands_wrappers: CommandWrapperCollection

    def __init__(
        self, app_path: str, cwd: str, executable_name: str, init_params: dict
    ):

        if "flags" in init_params:
            init_flags = init_params["flags"]
            if type(init_flags) == list:
                for flag in init_flags:
                    self._flags[flag] = flag

        self._working_directory = cwd
        self._user_home_directory = os.path.expanduser("~")
        self._working_directory_conf_path = os.path.join(
            self.get_working_directory(), self._config_folder_working_directory
        )
        self._config_folder_defaults_path = os.path.join(
            os.path.dirname(__file__), self._config_folder_defaults
        )
        self._user_home_directory_path = os.path.join(
            self._user_home_directory, self._config_folder_home_directory
        )
        self._cache_directory_path = os.path.join(
            self._user_home_directory_path, "cache"
        )
        self._config_folders = [
            self._config_folder_defaults_path,
            self._user_home_directory_path,
            self._working_directory_conf_path,
        ]

        self._config_object = CommandsConfig({})
        self._app_path = app_path
        self.init()
        self._executable_name = self._config_object.get(
            App.CONF_EXECUTABLE_NAME, executable_name
        )

    # <Init section
    def init(self):
        sys.path.append(self._working_directory_conf_path)
        sys.path.append(self._working_directory_conf_path)
        self.init_default_flags()
        self.init_logger()
        self.init_commands_configs()
        self._executable_path = self._config_object.get(App.CONF_EXECUTABLE_PATH, "")
        self.init_object_manager()

    def init_default_flags(self):
        to_remove_args = []
        for flag in sys.argv:
            if flag[0] == "-":
                if flag in self._available_flags:
                    self._flags[flag] = flag
                    to_remove_args.append(flag)
        for flag in to_remove_args:
            sys.argv.remove(flag)

        command_args = copy.deepcopy(sys.argv)
        command_args.pop(0)

        if len(command_args) > 0 and command_args[0] == "--":
            self._current_command = ""
        else:
            self._current_command = command_args

    def init_logger(self):
        import logging

        verb = logging.ERROR
        if self.is_verbose():
            verb = logging.INFO

        self.logger = setup_logger(
            "APP", verbosity=verb, file=self.get_log_file(self._log_file)
        )

    def init_commands_configs(self):
        config_object = self.construct_config_object(
            ConfigNamespaces.COMMANDS,
            self.get_config_files_by_namespace(ConfigNamespaces.COMMANDS),
            os.path.join(self._config_folder_defaults_path, "schema", "commands.json"),
        )
        self._config_object = CommandsConfig(config_object.get_all())
        self._commands = CommandCollection(self._config_object.get_commands())
        self._commands_wrappers = CommandWrapperCollection(
            self._config_object.get_commands_wrappers()
        )

    def get_config_files_by_namespace(self, namespace: str) -> list:
        config_files = []
        for folder in self._config_folders:
            if len(self._current_command) > 0:
                config_files.append(
                    os.path.join(folder, namespace, self._current_command[0] + ".yml")
                )
            else:
                for other_file in self.discover_configuration_file(
                    os.path.join(folder, namespace)
                ):
                    config_files.append(other_file)

            config_files.append(os.path.join(folder, namespace + ".yml"))

        return config_files

    @classmethod
    def discover_configuration_file(cls, directory: str):
        result = []
        if os.path.isdir(directory):
            for root, dirs, files in os.walk(directory):
                for file in files:
                    if file.endswith(".yml"):
                        result.append(os.path.join(directory, file))
        return result

    def construct_config_object(
        self,
        namespace: str,
        config_files: list,
        schema_file: str = "",
        first_file_mandatory: bool = False,
    ) -> Config:
        cache_keys = config_files.copy()
        cache_keys.append(namespace)
        cache_keys.append(self._working_directory_conf_path)
        cache = Cache(self._cache_directory_path, cache_keys, App.CACHE_LIFETIME)

        if not self.is_debug_enabled():
            try:
                return Config(cache.get())
            except CacheNotFoundException:
                pass

        try:
            config_reader = YamlConfigReader(
                config_files[0], required=first_file_mandatory, schema_file=schema_file
            )
            config_object = config_reader.get_config()
            for key in config_files[1:]:
                config_reader = YamlConfigReader(key, schema_file=schema_file)
                config_object.overwrite(config_reader.get_config())
        except ConfigValidationError as e:
            self.logger.error(e)
            self.output.print(
                "Configuration error: {0} at {1} in file {2}",
                [e.message, "/".join(e.error_path), e.file_path],
            )
            raise AppException(
                self.output.format(
                    "Unable to configure application. Fix errors in config file. {}",
                    [e.file_path],
                )
            )

        if not self.is_debug_enabled() and cache is not None:
            cache.save(config_object.get_all())

        return config_object

    @classmethod
    def get_args(cls):
        return sys.argv

    def get_config_object(self) -> CommandsConfig:
        return self._config_object

    def init_object_manager(self):
        self._object_manager = pinject.new_object_graph()

    def get_object_manager(self) -> pinject.object_graph.ObjectGraph:
        return self._object_manager

    # Init section>

    # < App Business logic section
    def run(self):
        if self.is_interactive():
            """Interactive shell for current command"""
            from lcli.app_mode.interactive import LcliPrompt

            LcliPrompt(self, " ".join(self._current_command)).run()
        else:
            """Fire application expose all commands"""
            from lcli.app_mode.fire import Fire

            Fire(self).run()

    # < App Business logic section

    # <Commands related methods
    def get_commands_wrappers(self) -> CommandWrapperCollection:
        return self._commands_wrappers

    def get_commands(self) -> CommandCollection:
        return self._commands

    def get_command_builder_factory(self):
        from lcli.command.builders import CommandBuilderFactory

        return CommandBuilderFactory(self)

    # Commands related methods>

    def get_app_code(self):
        return self._config_object.get("app_code")

    def get_app_description(self):
        return self._config_object.get("app_description", "")

    # <Flags
    def is_interactive(self):
        return self.FLAG_INTERACTIVE in self._flags

    def is_verbose(self):
        return self.FLAG_VERBOSE in self._flags

    def is_debug_enabled(self):
        return self.FLAG_DEBUG in self._flags

    # Flags>

    # <Directories/Paths
    def get_working_directory(self):
        return self._working_directory

    def get_executable_path(self):
        return self._executable_path

    def get_executable_name(self):
        return self._executable_name

    def get_self_executable_script(self):
        return os.path.join(self.get_executable_path(), self.get_executable_name())

    def get_log_file(self, log_file: str):
        log_path = os.path.join(self._user_home_directory_path, "log")
        try:
            os.makedirs(log_path)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        return os.path.join(log_path, log_file)

    # Directories/Paths>
    @property
    def cache_directory_path(self):
        return self._cache_directory_path
